@@ -8,26 +8,168 @@ import locale
 import copy
 import operator
 import warnings
+import jax
+from pint import Quantity, DimensionalityError
+from pint.util import SharedRegistryObject
 
 from .wrap_numpy import wrap_numpy, HANDLED_FUNCTIONS, HANDLED_UFUNCS
-from . import NegativeStdDevError
-from .util import is_np_duck_array
+from . import NegativeStdDevError, NumpyDowncastWarning
+from .util import is_np_duck_array, ignore_runtime_warnings, ignore_numpy_downcast_warnings, Display
 
 
-class Uncertainty(object):
+def _check_units(value, err):
+    mag_units = hasattr(value, "units")
+    err_units = hasattr(err, "units")
+    if mag_units ^ err_units and err is not None:
+        raise ValueError("Both mag and err need to have units if one of them has units!")
+    if mag_units and err is not None:
+        if value.units != err.units:
+            raise ValueError(
+                f"Value units {value.units} cannot be converted to error units {err.units}"
+            )
+    if mag_units:
+        mag_units = value.units
+        ret_val = value.to(mag_units).m
+        if err is not None:
+            ret_err = err.to(mag_units).m
+        else:
+            ret_err = None
+    else:
+        mag_units = 1.0
+        ret_val = value
+        ret_err = err
+
+    return ret_val, ret_err, mag_units
+
+
+def _strip_device_array(value, err):
+    if isinstance(value, jax.xla.DeviceArray):
+        value = value.to_py().copy()
+    if isinstance(err, jax.xla.DeviceArray):
+        err = err.to_py().copy()
+    return value, err
+
+
+class Uncertainty(Display):
     __apply_to_both_ndarray__ = ["flatten", "real", "imag", "astype", "T"]
     __ndarray_attributes__ = ["dtype", "ndim", "size"]
 
+    # Pint comparibility
+    @property
+    def unitless(self) -> bool:
+        """ """
+        return not bool(self.to_root_units()._units)
+
+    @property
+    def dimensionless(self) -> bool:
+        """ """
+        tmp = self.to_root_units()
+
+        return not bool(tmp.dimensionality)
+
+    _dimensionality = None
+
+    @property
+    def dimensionality(self):
+        """
+        Returns
+        -------
+        dict
+            Dimensionality of the Quantity, e.g. ``{length: 1, time: -1}``
+        """
+        if self._dimensionality is None:
+            self._dimensionality = self._REGISTRY._get_dimensionality(self.units)
+
+        return self._dimensionality
+
+    def check(self, dimension) -> bool:
+        """Return true if the quantity's dimension matches passed dimension."""
+        return self.dimensionality == self._REGISTRY.get_dimensionality(dimension)
+
+    def _check(self, other) -> bool:
+        """Check if the other object use a registry and if so that it is the
+        same registry.
+        Parameters
+        ----------
+        other :
+        Returns
+        -------
+        type
+            other don't use a registry and raise ValueError if other don't use the
+            same unit registry.
+        """
+        if self._REGISTRY is getattr(other, "_REGISTRY", None):
+            return True
+
+        elif isinstance(other, SharedRegistryObject):
+            mess = "Cannot operate with {} and {} of different registries."
+            raise ValueError(mess.format(self.__class__.__name__, other.__class__.__name__))
+        else:
+            return False
+
+    def to(self, other):
+        if hasattr(self.nominal_value, "units"):
+            return self.__class__(self._nom.to(other), self._err.to(other))
+        else:
+            raise AttributeError("Uncertainty has no quantity!")
+
+    @property
+    def m(self):
+        if hasattr(self.nominal_value, "units"):
+            return self.__class__(self._nom.m, self._err.m)
+        else:
+            raise AttributeError("Uncertainty has no quantity!")
+
+    @property
+    def units(self):
+        if hasattr(self.nominal_value, "units"):
+            return self._nom.units
+        else:
+            raise AttributeError("Uncertainty has no quantity!")
+
+    @property
+    def _REGISTRY(self):
+        return getattr(self._nom, "_REGISTRY", None)
+
+    def compatible_units(self, *contexts):
+        if contexts:
+            with self._REGISTRY.context(*contexts):
+                return self._REGISTRY.get_compatible_units(self._units)
+
+        return self._REGISTRY.get_compatible_units(self._units)
+
+    def _convert_magnitude_not_inplace(self, other, *contexts, **ctx_kwargs):
+        if contexts:
+            with self._REGISTRY.context(*contexts, **ctx_kwargs):
+                return self._REGISTRY.convert(self._magnitude, self._units, other)
+
+        return self._REGISTRY.convert(self._magnitude, self._units, other)
+
+    @ignore_numpy_downcast_warnings
     def __init__(self, value, err=None):
+
+        value, err, units = _check_units(value, err)
+        value, err = _strip_device_array(value, err)
+
+        # If Uncertatity
         if isinstance(value, self.__class__):
             magnitude_nom = value.value
             magnitude_err = value.error
+        # If sequence
+        elif isinstance(value, list):
+            return self.__class__.from_list(value)
+        # If arrays
         elif np.ndim(value) > 0:
             magnitude_nom = np.asarray(value)
             if err is None:
                 magnitude_err = np.zeros_like(value)
             else:
-                magnitude_err = np.asarray(err)
+                if np.ndim(err) == 0:
+                    magnitude_err = np.ones_like(value) * err
+                else:
+                    magnitude_err = np.asarray(err)
+                    assert magnitude_err.shape == magnitude_nom.shape
+        # If scalar
         else:
             magnitude_nom = value
             if err is None:
@@ -35,6 +177,14 @@ class Uncertainty(object):
             else:
                 magnitude_err = err
 
+        # Replace NaNs in errors with zeros
+        if is_np_duck_array(type(magnitude_err)):
+            magnitude_err[~np.isfinite(magnitude_err)] = 0
+        else:
+            if not np.isfinite(magnitude_err):
+                magnitude_err = 0
+        magnitude_nom *= units
+        magnitude_err *= units
         # Basic sanity checks
         if is_np_duck_array(type(magnitude_nom)):
             for item in self.__ndarray_attributes__ + ["shape"]:
@@ -42,22 +192,17 @@ class Uncertainty(object):
                     raise ValueError(
                         f"Attribute {item} does not match for value and error! ({getattr(magnitude_nom,item)} and {getattr(magnitude_err,item)})"
                     )
-        if np.any(np.atleast_1d(magnitude_err) < 0):
-            raise NegativeStdDevError
+        err_mag = np.atleast_1d(magnitude_err)
+        if np.any(err_mag[np.isfinite(err_mag)] < 0):
+            valid = err_mag[np.isfinite(err_mag)]
+
+            raise NegativeStdDevError(
+                f"Found negatives values for the standard deviation... {valid[valid < 0]}"
+            )
 
         self._nom = magnitude_nom
         self._err = magnitude_err
 
-    def __str__(self) -> str:
-        if self._nom is not None:
-            if self._err is not None:
-                return f"{self._nom} +/- {self._err}"
-            else:
-                return f"{self._nom}"
-
-    def __repr__(self) -> str:
-        return str(self)
-
     def __bytes__(self) -> bytes:
         return str(self).encode(locale.getpreferredencoding())
 
@@ -85,6 +230,10 @@ class Uncertainty(object):
     def nominal_value(self):
         return self.value
 
+    @property
+    def n(self):
+        return self.value
+
     @property
     def error(self):
         return self._err
@@ -93,12 +242,19 @@ class Uncertainty(object):
     def std_dev(self):
         return self.error
 
+    @property
+    def s(self):
+        return self.error
+
     @property
     def relative(self):
-        if self._nom != 0:
-            return self._err / self._nom
+        if np.ndim(self._nom) == 0:
+            if self._nom != 0:
+                return self._err / self._nom
+            else:
+                return np.NaN
         else:
-            return np.NaN
+            return self._err / self._nom
 
     @property
     def rel(self):
@@ -114,11 +270,18 @@ class Uncertainty(object):
         len_seq = len(seq)
         val = np.empty(len_seq)
         err = np.empty(len_seq)
-
-        for i, seq_i in enumerate(seq):
-            val[i] = seq_i._nom
-            err[i] = seq_i._err
-
+        if len_seq > 0:
+            first_item = seq[0]
+            if hasattr(first_item, "units"):
+                val *= first_item.units
+                err *= first_item.units
+            for i, seq_i in enumerate(seq):
+                try:
+                    val[i] = float(seq_i._nom)
+                    err[i] = float(seq_i._err)
+                except AttributeError:
+                    val[i] = float(seq_i)
+                    err[i] = 0
         return cls(val, err)
 
     def __float__(self) -> Uncertainty:
@@ -170,7 +333,7 @@ class Uncertainty(object):
         return self.__class__(new_mag, new_err)
 
     def __rsub__(self, other):
-        -self.__sub__(other)
+        return -self.__sub__(other)
 
     def __imul__(self, other):
         new = self * other
@@ -184,14 +347,16 @@ class Uncertainty(object):
     def __mul__(self, other):
         if isinstance(other, Uncertainty):
             new_mag = self._nom * other._nom
-            new_err = new_mag * np.sqrt(self.rel ** 2 + other.rel ** 2)
+            new_err = np.abs(new_mag) * np.sqrt(self.rel ** 2 + other.rel ** 2)
         else:
             new_mag = self._nom * other
-            new_err = self._err * other
+            new_err = np.abs(self._err * other)
+
         return self.__class__(new_mag, new_err)
 
     __rmul__ = __mul__
 
+    @ignore_runtime_warnings
     def __itruediv__(self, other):
         new = self / other
         if is_np_duck_array(type(self._nom)):
@@ -201,23 +366,25 @@ class Uncertainty(object):
         else:
             return new
 
+    @ignore_runtime_warnings
     def __truediv__(self, other):
         # Self / Other
         if isinstance(other, Uncertainty):
             new_mag = self._nom / other._nom
-            new_err = new_mag * np.sqrt(self.rel ** 2 + other.rel ** 2)
+            new_err = np.abs(new_mag) * np.sqrt(self.rel ** 2 + other.rel ** 2)
         else:
             new_mag = self._nom / other
-            new_err = self._err / other
+            new_err = np.abs(self._err / other)
         return self.__class__(new_mag, new_err)
 
+    @ignore_runtime_warnings
     def __rtruediv__(self, other):
         # Other / Self
         if isinstance(other, Uncertainty):
             raise Exception
         else:
             new_mag = other / self._nom
-            new_err = new_mag * self.rel
+            new_err = np.abs(new_mag) * np.abs(self.rel)
             return self.__class__(new_mag, new_err)
 
     __div__ = __truediv__
@@ -262,15 +429,20 @@ class Uncertainty(object):
     def __mod__(self, other):
         if isinstance(other, Uncertainty):
             new_mag = self._nom % other._nom
-            new_err = 0.0
         else:
             new_mag = self._nom % other
+        if np.ndim(new_mag) == 0:
             new_err = 0.0
+        else:
+            new_err = np.zeros_like(new_mag)
         return self.__class__(new_mag, new_err)
 
     def __rmod__(self, other):
         new_mag = other % self._nom
-        new_err = 0.0
+        if np.ndim(new_mag) == 0:
+            new_err = 0.0
+        else:
+            new_err = np.zeros_like(new_mag)
         return self.__class__(new_mag, new_err)
 
     def __divmod__(self, other):
@@ -288,20 +460,40 @@ class Uncertainty(object):
         else:
             return new
 
+    @ignore_runtime_warnings
     def __pow__(self, other):
         # Self ** other
+        A = self._nom
+        sA = self._err
         if isinstance(other, Uncertainty):
-            new_mag = self._nom % other._nom
-            new_err = 0.0
+            B = other._nom
+            sB = other._err
         else:
-            new_mag = self._nom % other
-            new_err = 0.0
+            B = other
+            sB = 0
+        new_mag = A ** B
+        if sB == 0 and int(B) == B:
+            new_err = np.abs(new_mag) * np.sqrt((B / A * sA) ** 2)
+        else:
+            new_err = np.sqrt((B / A * sA) ** 2 + (np.log(A) * sB) ** 2)
+
         return self.__class__(new_mag, new_err)
 
+    @ignore_runtime_warnings
     def __rpow__(self, other):
         # Other ** self
-        new_mag = other % self._nom
-        new_err = 0.0
+        B = self._nom
+        sB = self._err
+        if isinstance(other, Uncertainty):
+            A = other._nom
+            sA = other._err
+        else:
+            A = other
+            sA = 0
+
+        new_mag = A ** B
+        new_err = np.abs(new_mag) * np.sqrt((B / A * sA) ** 2 + (np.log(A) * sB) ** 2)
+
         return self.__class__(new_mag, new_err)
 
     def __abs__(self):
@@ -349,6 +541,7 @@ class Uncertainty(object):
     # NumPy function/ufunc support
     __array_priority__ = 17
 
+    @ignore_runtime_warnings
     def __array_function__(self, func, types, args, kwargs):
         # print(func)
         if func.__name__ not in HANDLED_FUNCTIONS:
@@ -358,13 +551,14 @@ class Uncertainty(object):
         else:
             return wrap_numpy("function", func, args, kwargs)
 
+    @ignore_runtime_warnings
     def __array_ufunc__(self, ufunc, method, *args, **kwargs):
         # print(method,ufunc.__name__)
         if method != "__call__":
             raise NotImplementedError
         else:
             if ufunc.__name__ not in HANDLED_UFUNCS:
-                raise NotImplementedError
+                raise NotImplementedError(f"Ufunc {ufunc.__name__} is not implemented!") from None
             else:
                 return wrap_numpy("ufunc", ufunc, args, kwargs)
 
@@ -388,12 +582,41 @@ class Uncertainty(object):
             return lambda *args, **kwargs: wrap_numpy("function", item, [self] + list(args), kwargs)
         elif item in self.__ndarray_attributes__:
             return getattr(self._nom, item)
+        elif hasattr(Quantity, item):
+            val = getattr(self._nom, item)
+            err = getattr(self._err, item)
+            if callable(val):
+
+                def expr(*args, **kwargs):
+                    try:
+                        vexpr = val(*args, **kwargs)
+                        eexpr = err(*args, **kwargs)
+                    except Exception:
+                        raise ValueError(
+                            f"Could not execute method {item} on Uncertainty elements!"
+                        )
+                    try:
+                        ret_instance = self.__class__(vexpr, eexpr)
+                    except Exception:
+                        raise ValueError(
+                            f"Could not execute instantiate Uncertainty class with results from method {item}!"
+                        )
+                    else:
+                        return ret_instance
+
+                return expr
+            else:
+                return self.__class__(val, err)
         else:
-            raise AttributeError
+            raise AttributeError(
+                f"Attribute {item} not available in Uncertainty, as method of a Pint Quantity, or as NumPy ufunc or function."
+            ) from None
 
     def __array__(self, t=None) -> np.ndarray:
         warnings.warn(
-            "The uncertainty is stripped when downcasting to ndarray.", UserWarning, stacklevel=2
+            "The uncertainty is stripped when downcasting to ndarray.",
+            NumpyDowncastWarning,
+            stacklevel=2,
         )
         return np.asarray(self._nom)
 
@@ -411,6 +634,10 @@ class Uncertainty(object):
         else:
             raise ValueError("Can only 'put' Uncertainties into uncertainties!")
 
+    def copy(self):
+        return Uncertainty(self._nom.copy(), self._err.copy())
+
+    # Special properties
     @property
     def flat(self):
         for u, v in (self._nom.flat, self._err.flat):
@@ -425,6 +652,10 @@ class Uncertainty(object):
         self._nom.shape = value
         self._err.shape = value
 
+    @property
+    def nbytes(self):
+        return self._nom.nbytes + self._err.nbytes
+
     def searchsorted(self, v, side="left", sorter=None):
         return self._nom.searchsorted(v, side)
 
